# # async def run_test():
#
# if __name__ == "__main__":
#     logging.basicConfig(level=logging.DEBUG)
#
#     iface = modbus_iface()
#     iface.open_bus(
#         bus_type="serial",
#         name="test",
#         serial_port="/dev/ttyk37simOut",
#         serial_baud=38400,
#         serial_method="rtu",
#         serial_bits=8,
#         serial_parity="N",
#         serial_stop=1,
#         serial_timeout=0.3,
#     )
#
#     print(iface.getBusStatus(bus_id="test"))
#
#     result = iface.read_registers(
#         bus_id="test",
#         modbus_id=1,
#         start_address=0,
#         num_registers=23,
#     )
#     print(result)
#
#     watchdog = result[22] + 1
#
#     result = iface.write_registers(
#         bus_id="test",
#         modbus_id=1,
#         start_address=22,
#         values=[watchdog],
#     )
#
#     ## define a function to print the results of read register subscription
#     def print_results(values):
#         print(values)
#
#     loop = asyncio.get_event_loop()
#
#     ## add a read register subscription
#     iface.add_read_register_subscription(
#         bus_id="test",
#         modbus_id=1,
#         start_address=0,
#         num_registers=23,
#         callback=print_results,
#     )
#
#     print("Subscribed to read register subscription")
#
#     ## add a read register subscription
#     iface.add_read_register_subscription(
#         bus_id="test",
#         modbus_id=1,
#         start_address=0,
#         num_registers=10,
#         callback=print_results,
#     )
#
#     # async def run_test():
#     #     await asyncio.sleep(20)
#     #     iface.close()
#
#     try:
#         asyncio.get_event_loop().run_forever()
#     except KeyboardInterrupt:
#         pass
#
#     log.info("Closing modbus interface")
#     # iface.close()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: grpc_stubs/modbus_iface.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dgrpc_stubs/modbus_iface.proto\x12\x0cmodbus_iface\"\xd5\x01\n\tbusStatus\x12\x0e\n\x06\x62us_id\x18\x01 \x01(\t\x12\x0c\n\x04open\x18\x02 \x01(\x08\x12=\n\x0fserial_settings\x18\x03 \x01(\x0b\x32\x1f.modbus_iface.serialBusSettingsH\x00\x88\x01\x01\x12\x41\n\x11\x65thernet_settings\x18\x04 \x01(\x0b\x32!.modbus_iface.ethernetBusSettingsH\x01\x88\x01\x01\x42\x12\n\x10_serial_settingsB\x14\n\x12_ethernet_settings\"\xb0\x01\n\x0eresponseHeader\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x1a\n\rresponse_code\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x1d\n\x10response_message\x18\x03 \x01(\tH\x01\x88\x01\x01\x12+\n\nbus_status\x18\x04 \x03(\x0b\x32\x17.modbus_iface.busStatusB\x10\n\x0e_response_codeB\x13\n\x11_response_message\"\x89\x01\n\x14serverResponseHeader\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x1a\n\rresponse_code\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x1d\n\x10response_message\x18\x03 \x01(\tH\x01\x88\x01\x01\x42\x10\n\x0e_response_codeB\x13\n\x11_response_message\"#\n\x10testCommsRequest\x12\x0f\n\x07message\x18\x01 \x01(\t\"\\\n\x11testCommsResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.modbus_iface.responseHeader\x12\x10\n\x08response\x18\x02 \x01(\t\"\xeb\x01\n\x11serialBusSettings\x12\x0c\n\x04port\x18\x01 \x01(\t\x12\x0c\n\x04\x62\x61ud\x18\x02 \x01(\x05\x12\x1a\n\rmodbus_method\x18\x03 \x01(\tH\x00\x88\x01\x01\x12\x16\n\tdata_bits\x18\x04 \x01(\x05H\x01\x88\x01\x01\x12\x13\n\x06parity\x18\x05 \x01(\tH\x02\x88\x01\x01\x12\x16\n\tstop_bits\x18\x06 \x01(\x05H\x03\x88\x01\x01\x12\x14\n\x07timeout\x18\x07 \x01(\x02H\x04\x88\x01\x01\x42\x10\n\x0e_modbus_methodB\x0c\n\n_data_bitsB\t\n\x07_parityB\x0c\n\n_stop_bitsB\n\n\x08_timeout\"\x7f\n\x13\x65thernetBusSettings\x12\n\n\x02ip\x18\x01 \x01(\t\x12\x0c\n\x04port\x18\x02 \x01(\x05\x12\x1a\n\rmodbus_method\x18\x03 \x01(\tH\x00\x88\x01\x01\x12\x14\n\x07timeout\x18\x04 \x01(\x02H\x01\x88\x01\x01\x42\x10\n\x0e_modbus_methodB\n\n\x08_timeout\"\xac\x01\n\x0eopenBusRequest\x12\x0e\n\x06\x62us_id\x18\x01 \x01(\t\x12:\n\x0fserial_settings\x18\x02 \x01(\x0b\x32\x1f.modbus_iface.serialBusSettingsH\x00\x12>\n\x11\x65thernet_settings\x18\x03 \x01(\x0b\x32!.modbus_iface.ethernetBusSettingsH\x00\x42\x0e\n\x0c\x62us_settings\"X\n\x0fopenBusResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.modbus_iface.responseHeader\x12\x0e\n\x06\x62us_id\x18\x02 \x01(\t\"\x10\n\x0elistBusRequest\"u\n\x0flistBusResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.modbus_iface.responseHeader\x12+\n\nbus_status\x18\x02 \x03(\x0b\x32\x17.modbus_iface.busStatus\"\"\n\x10\x62usStatusRequest\x12\x0e\n\x06\x62us_id\x18\x01 \x01(\t\"w\n\x11\x62usStatusResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.modbus_iface.responseHeader\x12+\n\nbus_status\x18\x02 \x01(\x0b\x32\x17.modbus_iface.busStatus\"!\n\x0f\x63loseBusRequest\x12\x0e\n\x06\x62us_id\x18\x01 \x01(\t\"Y\n\x10\x63loseBusResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.modbus_iface.responseHeader\x12\x0e\n\x06\x62us_id\x18\x02 \x01(\t\"o\n\x13readRegisterRequest\x12\x0e\n\x06\x62us_id\x18\x01 \x01(\t\x12\x11\n\tmodbus_id\x18\x02 \x01(\x05\x12\x15\n\rregister_type\x18\x03 \x01(\x05\x12\x0f\n\x07\x61\x64\x64ress\x18\x04 \x01(\x05\x12\r\n\x05\x63ount\x18\x05 \x01(\x05\"t\n\x14readRegisterResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.modbus_iface.responseHeader\x12\x15\n\rresponse_code\x18\x02 \x01(\x05\x12\x0e\n\x06values\x18\x03 \x03(\x05\"q\n\x14writeRegisterRequest\x12\x0e\n\x06\x62us_id\x18\x01 \x01(\t\x12\x11\n\tmodbus_id\x18\x02 \x01(\x05\x12\x15\n\rregister_type\x18\x03 \x01(\x05\x12\x0f\n\x07\x61\x64\x64ress\x18\x04 \x01(\x05\x12\x0e\n\x06values\x18\x05 \x03(\x05\"e\n\x15writeRegisterResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.modbus_iface.responseHeader\x12\x15\n\rresponse_code\x18\x02 \x01(\x05\"z\n\x1cscheduleWriteRegisterRequest\x12\x0e\n\x06\x62us_id\x18\x01 \x01(\t\x12\x15\n\rregister_type\x18\x02 \x01(\x05\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\x05\x12\x0e\n\x06values\x18\x04 \x03(\x05\x12\x12\n\ndelay_secs\x18\x05 \x01(\x05\"z\n\x1dscheduleWriteRegisterResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.modbus_iface.responseHeader\x12\x0e\n\x06\x62us_id\x18\x02 \x01(\t\x12\x12\n\ndelay_secs\x18\x03 \x01(\x05\"\x8e\x01\n\x1freadRegisterSubscriptionRequest\x12\x0e\n\x06\x62us_id\x18\x01 \x01(\t\x12\x11\n\tmodbus_id\x18\x02 \x01(\x05\x12\x15\n\rregister_type\x18\x03 \x01(\x05\x12\x0f\n\x07\x61\x64\x64ress\x18\x04 \x01(\x05\x12\r\n\x05\x63ount\x18\x05 \x01(\x05\x12\x11\n\tpoll_secs\x18\x06 \x01(\x05\"\x90\x01\n readRegisterSubscriptionResponse\x12\x35\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\x1c.modbus_iface.responseHeader\x12\x0e\n\x06\x62us_id\x18\x02 \x01(\t\x12\x15\n\rresponse_code\x18\x03 \x01(\x05\x12\x0e\n\x06values\x18\x04 \x03(\x05\"\xca\x01\n\x13\x63reateServerRequest\x12\x0c\n\x04type\x18\x01 \x01(\t\x12\x11\n\x04port\x18\x02 \x01(\x05H\x00\x88\x01\x01\x12\x11\n\x04host\x18\x03 \x01(\tH\x01\x88\x01\x01\x12\x19\n\x11holding_registers\x18\x04 \x03(\x05\x12\x17\n\x0finput_registers\x18\x05 \x03(\x05\x12\r\n\x05\x63oils\x18\x06 \x03(\x08\x12\x17\n\x0f\x64iscrete_inputs\x18\x07 \x03(\x08\x12\x11\n\tmodbus_id\x18\x08 \x01(\x05\x42\x07\n\x05_portB\x07\n\x05_host\"e\n\x14\x63reateServerResponse\x12;\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\".modbus_iface.serverResponseHeader\x12\x10\n\x08response\x18\x02 \x01(\t\"\x13\n\x11listServerRequest\"b\n\x12listServerResponse\x12;\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\".modbus_iface.serverResponseHeader\x12\x0f\n\x07servers\x18\x02 \x03(\t\"\"\n\x12\x63loseServerRequest\x12\x0c\n\x04port\x18\x01 \x01(\x05\"`\n\x13\x63loseServerResponse\x12;\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\".modbus_iface.serverResponseHeader\x12\x0c\n\x04port\x18\x02 \x01(\x05\"a\n\x19setServerRegistersRequest\x12\x0c\n\x04port\x18\x01 \x01(\x05\x12\x15\n\rfunction_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x03(\x05\x12\x0e\n\x06values\x18\x04 \x03(\x05\"Y\n\x1asetServerRegistersResponse\x12;\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\".modbus_iface.serverResponseHeader\"`\n\x19getServerRegistersRequest\x12\x0c\n\x04port\x18\x01 \x01(\x05\x12\x15\n\rfunction_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\x05\x12\r\n\x05\x63ount\x18\x04 \x01(\x05\"i\n\x1agetServerRegistersResponse\x12;\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\".modbus_iface.serverResponseHeader\x12\x0e\n\x06values\x18\x02 \x03(\x05\"z\n\x1escheduleServerRegistersRequest\x12\x0c\n\x04port\x18\x01 \x01(\x05\x12\x15\n\rfunction_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\x05\x12\x0e\n\x06values\x18\x04 \x03(\x05\x12\x12\n\ndelay_secs\x18\x05 \x01(\x02\"^\n\x1fscheduleServerRegistersResponse\x12;\n\x0fresponse_header\x18\x01 \x01(\x0b\x32\".modbus_iface.serverResponseHeader2\x8b\x0b\n\x0bmodbusIface\x12N\n\ttestComms\x12\x1e.modbus_iface.testCommsRequest\x1a\x1f.modbus_iface.testCommsResponse\"\x00\x12H\n\x07openBus\x12\x1c.modbus_iface.openBusRequest\x1a\x1d.modbus_iface.openBusResponse\"\x00\x12H\n\x07listBus\x12\x1c.modbus_iface.listBusRequest\x1a\x1d.modbus_iface.listBusResponse\"\x00\x12N\n\tbusStatus\x12\x1e.modbus_iface.busStatusRequest\x1a\x1f.modbus_iface.busStatusResponse\"\x00\x12K\n\x08\x63loseBus\x12\x1d.modbus_iface.closeBusRequest\x1a\x1e.modbus_iface.closeBusResponse\"\x00\x12X\n\rreadRegisters\x12!.modbus_iface.readRegisterRequest\x1a\".modbus_iface.readRegisterResponse\"\x00\x12[\n\x0ewriteRegisters\x12\".modbus_iface.writeRegisterRequest\x1a#.modbus_iface.writeRegisterResponse\"\x00\x12s\n\x16scheduleWriteRegisters\x12*.modbus_iface.scheduleWriteRegisterRequest\x1a+.modbus_iface.scheduleWriteRegisterResponse\"\x00\x12}\n\x18readRegisterSubscription\x12-.modbus_iface.readRegisterSubscriptionRequest\x1a..modbus_iface.readRegisterSubscriptionResponse\"\x00\x30\x01\x12W\n\x0c\x63reateServer\x12!.modbus_iface.createServerRequest\x1a\".modbus_iface.createServerResponse\"\x00\x12Q\n\nlistServer\x12\x1f.modbus_iface.listServerRequest\x1a .modbus_iface.listServerResponse\"\x00\x12T\n\x0b\x63loseServer\x12 .modbus_iface.closeServerRequest\x1a!.modbus_iface.closeServerResponse\"\x00\x12i\n\x12setServerRegisters\x12\'.modbus_iface.setServerRegistersRequest\x1a(.modbus_iface.setServerRegistersResponse\"\x00\x12i\n\x12getServerRegisters\x12\'.modbus_iface.getServerRegistersRequest\x1a(.modbus_iface.getServerRegistersResponse\"\x00\x12x\n\x17scheduleServerRegisters\x12,.modbus_iface.scheduleServerRegistersRequest\x1a-.modbus_iface.scheduleServerRegistersResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'grpc_stubs.modbus_iface_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_BUSSTATUS']._serialized_start=48
  _globals['_BUSSTATUS']._serialized_end=261
  _globals['_RESPONSEHEADER']._serialized_start=264
  _globals['_RESPONSEHEADER']._serialized_end=440
  _globals['_SERVERRESPONSEHEADER']._serialized_start=443
  _globals['_SERVERRESPONSEHEADER']._serialized_end=580
  _globals['_TESTCOMMSREQUEST']._serialized_start=582
  _globals['_TESTCOMMSREQUEST']._serialized_end=617
  _globals['_TESTCOMMSRESPONSE']._serialized_start=619
  _globals['_TESTCOMMSRESPONSE']._serialized_end=711
  _globals['_SERIALBUSSETTINGS']._serialized_start=714
  _globals['_SERIALBUSSETTINGS']._serialized_end=949
  _globals['_ETHERNETBUSSETTINGS']._serialized_start=951
  _globals['_ETHERNETBUSSETTINGS']._serialized_end=1078
  _globals['_OPENBUSREQUEST']._serialized_start=1081
  _globals['_OPENBUSREQUEST']._serialized_end=1253
  _globals['_OPENBUSRESPONSE']._serialized_start=1255
  _globals['_OPENBUSRESPONSE']._serialized_end=1343
  _globals['_LISTBUSREQUEST']._serialized_start=1345
  _globals['_LISTBUSREQUEST']._serialized_end=1361
  _globals['_LISTBUSRESPONSE']._serialized_start=1363
  _globals['_LISTBUSRESPONSE']._serialized_end=1480
  _globals['_BUSSTATUSREQUEST']._serialized_start=1482
  _globals['_BUSSTATUSREQUEST']._serialized_end=1516
  _globals['_BUSSTATUSRESPONSE']._serialized_start=1518
  _globals['_BUSSTATUSRESPONSE']._serialized_end=1637
  _globals['_CLOSEBUSREQUEST']._serialized_start=1639
  _globals['_CLOSEBUSREQUEST']._serialized_end=1672
  _globals['_CLOSEBUSRESPONSE']._serialized_start=1674
  _globals['_CLOSEBUSRESPONSE']._serialized_end=1763
  _globals['_READREGISTERREQUEST']._serialized_start=1765
  _globals['_READREGISTERREQUEST']._serialized_end=1876
  _globals['_READREGISTERRESPONSE']._serialized_start=1878
  _globals['_READREGISTERRESPONSE']._serialized_end=1994
  _globals['_WRITEREGISTERREQUEST']._serialized_start=1996
  _globals['_WRITEREGISTERREQUEST']._serialized_end=2109
  _globals['_WRITEREGISTERRESPONSE']._serialized_start=2111
  _globals['_WRITEREGISTERRESPONSE']._serialized_end=2212
  _globals['_SCHEDULEWRITEREGISTERREQUEST']._serialized_start=2214
  _globals['_SCHEDULEWRITEREGISTERREQUEST']._serialized_end=2336
  _globals['_SCHEDULEWRITEREGISTERRESPONSE']._serialized_start=2338
  _globals['_SCHEDULEWRITEREGISTERRESPONSE']._serialized_end=2460
  _globals['_READREGISTERSUBSCRIPTIONREQUEST']._serialized_start=2463
  _globals['_READREGISTERSUBSCRIPTIONREQUEST']._serialized_end=2605
  _globals['_READREGISTERSUBSCRIPTIONRESPONSE']._serialized_start=2608
  _globals['_READREGISTERSUBSCRIPTIONRESPONSE']._serialized_end=2752
  _globals['_CREATESERVERREQUEST']._serialized_start=2755
  _globals['_CREATESERVERREQUEST']._serialized_end=2957
  _globals['_CREATESERVERRESPONSE']._serialized_start=2959
  _globals['_CREATESERVERRESPONSE']._serialized_end=3060
  _globals['_LISTSERVERREQUEST']._serialized_start=3062
  _globals['_LISTSERVERREQUEST']._serialized_end=3081
  _globals['_LISTSERVERRESPONSE']._serialized_start=3083
  _globals['_LISTSERVERRESPONSE']._serialized_end=3181
  _globals['_CLOSESERVERREQUEST']._serialized_start=3183
  _globals['_CLOSESERVERREQUEST']._serialized_end=3217
  _globals['_CLOSESERVERRESPONSE']._serialized_start=3219
  _globals['_CLOSESERVERRESPONSE']._serialized_end=3315
  _globals['_SETSERVERREGISTERSREQUEST']._serialized_start=3317
  _globals['_SETSERVERREGISTERSREQUEST']._serialized_end=3414
  _globals['_SETSERVERREGISTERSRESPONSE']._serialized_start=3416
  _globals['_SETSERVERREGISTERSRESPONSE']._serialized_end=3505
  _globals['_GETSERVERREGISTERSREQUEST']._serialized_start=3507
  _globals['_GETSERVERREGISTERSREQUEST']._serialized_end=3603
  _globals['_GETSERVERREGISTERSRESPONSE']._serialized_start=3605
  _globals['_GETSERVERREGISTERSRESPONSE']._serialized_end=3710
  _globals['_SCHEDULESERVERREGISTERSREQUEST']._serialized_start=3712
  _globals['_SCHEDULESERVERREGISTERSREQUEST']._serialized_end=3834
  _globals['_SCHEDULESERVERREGISTERSRESPONSE']._serialized_start=3836
  _globals['_SCHEDULESERVERREGISTERSRESPONSE']._serialized_end=3930
  _globals['_MODBUSIFACE']._serialized_start=3933
  _globals['_MODBUSIFACE']._serialized_end=5352
# @@protoc_insertion_point(module_scope)



#!/usr/bin/env python3

import time
import logging


## A simple 1D Kalman filter implementation
## This filter is designed to be used with a single sensor reading (e.g., a voltage reading, a temperature reading, etc.)
## It includes automatic time step calculation based on the time since the last update, however this can be overridden by providing a dt parameter.
## Has some basic outlier detection and handling

## If you want to adjust the overall sensitivity of the filter, you can adjust the process_variance parameter.
## Default is 0.5, which is a good starting point for most applications. 1 will make the filter more sensitive to changes, 0.1 will make it less sensitive.

## If you want to adjust the sensitivity to outliers, you can adjust the outlier_threshold which defaults to 5.
## Increase this value to make the filter less sensitive to outliers.

## If you are not familiar with Kalman filters, here is a good resource:
## https://www.bzarg.com/p/how-a-kalman-filter-works-in-pictures/


class KalmanFilter1D:
    def __init__(
        self,
        initial_estimate=None,
        initial_error_estimate=None,
        process_variance=None,
        outlier_protection=None,
        outlier_threshold=None,
        outlier_variance_multiplier=None,
    ):
        self.debug = False

        self.default_process_variance = 0.5  # Default process variance if not provided
        self.default_initial_estimate_ratio = 25  # The ratio of the initial estimate variance to the first measurement variance
        self.default_measurement_variance = (
            0.5  # Default measurement variance if not provided
        )

        self.outlier_protection = (
            outlier_protection if outlier_protection is not None else True
        )  # Whether to use outlier protection
        self.default_outlier_threshold = (
            5  # The threshold for an outlier (in multiples of the error estimate)
        )
        self.default_outlier_variance_multiplier = (
            25  # Multiplier for the measurement variance for an outlier
        )

        self.estimate = initial_estimate  # Initial estimate (x_hat)
        self.error_estimate = initial_error_estimate  # Initial error estimate (P)
        self.process_variance = (
            process_variance or self.default_process_variance
        )  # Process variance (Q)

        self.outlier_threshold = (
            outlier_threshold or self.default_outlier_threshold
        )  # Threshold for an outlier
        self.outlier_variance_multiplier = (
            outlier_variance_multiplier or self.default_outlier_variance_multiplier
        )

        self.kalman_gain = 0  # Kalman gain (K)
        self.last_timestamp = None  # To store the last time the filter was updated
        self.enabled = True

    def set_estimate(self, estimate):
        self.estimate = estimate

    def set_error_estimate(self, error_estimate):
        self.error_estimate = error_estimate

    def set_process_variance(self, process_variance):
        self.process_variance = process_variance

    def set_outlier_protection(self, outlier_protection):
        self.outlier_protection = outlier_protection

    def set_outlier_threshold(self, outlier_threshold):
        self.outlier_threshold = outlier_threshold

    def set_outlier_variance_multiplier(self, outlier_variance_multiplier):
        self.outlier_variance_multiplier = outlier_variance_multiplier

    def update(
        self,
        measurement,
        measurement_variance=None,
        dt=None,
        outlier_protection=None,
        process_variance=None,
    ):
        if self.enabled is False:
            # allow a way of disabling kalman filter in test mode.
            return measurement

        if process_variance is not None:
            if self.debug:
                logging.debug(f"Setting process variance to {process_variance}")
            self.process_variance = process_variance

        ## If the measurement is None, return the last estimate
        if measurement is None:
            return self.estimate

        ## If the measurement variance is not provided, set it to the default
        measurement_variance = measurement_variance or self.default_measurement_variance

        ## If not initialized, set the initial estimate to the first measurement
        if self.error_estimate is None:
            if self.debug:
                logging.debug("Initializing Kalman filter")
            self.error_estimate = (
                measurement_variance * self.default_initial_estimate_ratio
            )
        if self.estimate is None:
            self.estimate = measurement
            self.last_timestamp = time.time()
            return self.estimate

        current_time = time.time()
        # If dt is not provided, calculate it based on the time since the last update
        if dt is None:
            if self.last_timestamp is None:
                dt = 1  # Default to 1 second if this is the first call
            else:
                dt = current_time - self.last_timestamp
        self.last_timestamp = current_time

        # If the measurement is an outlier, increase the measurement variance
        outlier_protection = (
            outlier_protection
            if outlier_protection is not None
            else self.outlier_protection
        )
        outlier_message = None
        if (
            outlier_protection
            and abs(measurement - self.estimate)
            > self.outlier_threshold * self.error_estimate
        ):
            # logging.debug(f"Outlier detected: {measurement} (threshold: {self.outlier_threshold * self.error_estimate})")
            outlier_message = f"Outlier detected: {measurement} (threshold: {self.outlier_threshold * self.error_estimate})"
            measurement_variance *= self.outlier_variance_multiplier

        # Adjust process variance with dt
        adjusted_process_variance = self.process_variance * dt

        # Also adjust the process variance based on the magnitude of the measurement. This makes the filter more performant over datasets of wider magnitude.
        # adjusted_process_variance = self.process_variance * (1 + (abs(measurement) / 10)) ## This is a good starting point for most applications

        # Prediction step
        self.error_estimate += adjusted_process_variance

        # Update step
        if (self.error_estimate + measurement_variance) == 0:
            ## If the error estimate + measurement variance is 0, avoid division by zero
            self.kalman_gain = self.error_estimate / 0.0001
        else:
            self.kalman_gain = self.error_estimate / (
                self.error_estimate + measurement_variance
            )
        self.estimate += self.kalman_gain * (measurement - self.estimate)
        self.error_estimate = (1 - self.kalman_gain) * self.error_estimate

        debug_output = f"Measurement: {measurement}, Estimate: {self.estimate}, Error estimate: {self.error_estimate}"
        if outlier_message:
            debug_output += f" | {outlier_message}"
        if self.debug:
            logging.debug(debug_output)

        return self.estimate


def apply_kalman_filter(
    initial_estimate=None,
    initial_error_estimate=None,
    process_variance=None,
    outlier_protection=None,
    outlier_threshold=None,
    outlier_variance_multiplier=None,
):
    """A decorator to apply a Kalman filter to the return value of a function
    The function should return a single value (e.g., a sensor reading)
    See below for an example of how to use this decorator

    Parameters
    ----------
    initial_estimate
    initial_error_estimate
    process_variance
    outlier_protection
    outlier_threshold
    outlier_variance_multiplier

    Returns
    -------

    """

    def decorator(func):
        ## Generate an id for the Kalman filter instance
        kf_id = id(func)

        def wrapper(self, *args, **kwargs):
            ## Instantiate the Kalman filter if it doesn't exist
            ## This allows multiple instances of the same function to have separate Kalman filters
            if not hasattr(self, "_kf_instances"):
                self._kf_instances = {}
            if kf_id not in self._kf_instances:
                # Create an instance of the Kalman filter with the provided parameters
                self._kf_instances[kf_id] = KalmanFilter1D(
                    initial_estimate=initial_estimate,
                    initial_error_estimate=initial_error_estimate,
                    process_variance=process_variance,
                    outlier_protection=outlier_protection,
                    outlier_threshold=outlier_threshold,
                    outlier_variance_multiplier=outlier_variance_multiplier,
                )

            _kalman_filter = self._kf_instances[kf_id]

            wrapper._kalman_filter = _kalman_filter

            # Call the original function and get its return value
            result = func(self, *args, **kwargs)

            # Extract measurement_variance and optional dt from kwargs or set defaults
            measurement_variance = kwargs.pop("kf_measurement_variance", None)
            dt = kwargs.pop("kf_dt", None)  # Allow dt to be optionally overridden
            measurement_outlier_protection = kwargs.pop(
                "kf_outlier_protection", None
            )  # Allow outlier_protection to be optionally overridden
            kf_process_variance = kwargs.pop(
                "kf_process_variance", None
            )  # Allow process_variance to be optionally overridden

            # Apply the Kalman filter to the return value
            filtered_result = _kalman_filter.update(
                result,
                measurement_variance,
                dt,
                measurement_outlier_protection,
                kf_process_variance,
            )
            return filtered_result

        return wrapper

    return decorator


def apply_async_kalman_filter(
    initial_estimate=None,
    initial_error_estimate=None,
    process_variance=None,
    outlier_protection=None,
    outlier_threshold=None,
    outlier_variance_multiplier=None,
):
    def decorator(func):
        ## Generate an id for the Kalman filter instance
        kf_id = id(func)

        async def wrapper(self, *args, **kwargs):
            ## Instantiate the Kalman filter if it doesn't exist
            ## This allows multiple instances of the same function to have separate Kalman filters
            if not hasattr(self, "_kf_instances"):
                self._kf_instances = {}
            if kf_id not in self._kf_instances:
                # Create an instance of the Kalman filter with the provided parameters
                self._kf_instances[kf_id] = KalmanFilter1D(
                    initial_estimate=initial_estimate,
                    initial_error_estimate=initial_error_estimate,
                    process_variance=process_variance,
                    outlier_protection=outlier_protection,
                    outlier_threshold=outlier_threshold,
                    outlier_variance_multiplier=outlier_variance_multiplier,
                )

            _kalman_filter = self._kf_instances[kf_id]

            wrapper._kalman_filter = _kalman_filter

            # Call the original function and get its return value
            result = await func(self, *args, **kwargs)

            # Extract measurement_variance and optional dt from kwargs or set defaults
            measurement_variance = kwargs.pop("kf_measurement_variance", None)
            dt = kwargs.pop("kf_dt", None)  # Allow dt to be optionally overridden
            measurement_outlier_protection = kwargs.pop(
                "kf_outlier_protection", None
            )  # Allow outlier_protection to be optionally overridden
            kf_process_variance = kwargs.pop(
                "kf_process_variance", None
            )  # Allow process_variance to be optionally overridden

            # Apply the Kalman filter to the return value
            filtered_result = _kalman_filter.update(
                result,
                measurement_variance,
                dt,
                measurement_outlier_protection,
                kf_process_variance,
            )
            return filtered_result

        return wrapper

    return decorator



"""
flatten() and unflatten()

A pair of functions that can convert an arbitrary JSON object into a
flat name/value pair dictionary and back again, preserving type
information and handling both nested lists and nested dictionaries.

For example:

    {
        "this": {
            "is": {
                "nested": [{
                    "nested_dict_one": 10
                }, {
                    "nested_dict_two": 20.5
                }]
            },
            "other_types": {
                "true": True,
                "false": False,
                "none": None,
            }
        }
    }

Flattens to:

    {
        "this.is.nested.[0].nested_dict_one$int": "10",
        "this.is.nested.[1].nested_dict_two$float": "20.5",
        "this.other_types.true$bool": "True",
        "this.other_types.false$bool": "False",
        "this.other_types.none$none": "None",
    }
"""

import re


def _object_to_rows(obj, prefix=None):
    rows = []
    dot_prefix = prefix and (prefix + ".") or ""
    if isinstance(obj, dict):
        if not obj:
            rows.append(((prefix or "") + "$empty", "{}"))
        else:
            for key, item in obj.items():
                rows.extend(_object_to_rows(item, prefix=dot_prefix + key))
    elif isinstance(obj, (list, tuple)):
        if len(obj) == 0:
            rows.append(((prefix or "") + "$emptylist", "[]"))
        for i, item in enumerate(obj):
            rows.extend(_object_to_rows(item, prefix=dot_prefix + "[{}]".format(i)))
    elif obj is None:
        rows.append(((prefix or "") + "$none", "None"))
    elif isinstance(obj, bool):
        rows.append(((prefix or "") + "$bool", str(obj)))
    elif isinstance(obj, int):
        rows.append(((prefix or "") + "$int", str(obj)))
    elif isinstance(obj, float):
        rows.append(((prefix or "") + "$float", str(obj)))
    else:
        rows.append((prefix, str(obj)))
    return rows


def flatten(obj):
    return dict(_object_to_rows(obj))


_types_re = re.compile(r".*\$(none|bool|int|float|empty|emptylist)$")
_int_key_re = re.compile(r"\[(\d+)\]")


def unflatten(data):
    obj = {}
    for key, value in data.items():
        current = obj
        bits = key.split(".")
        path, lastkey = bits[:-1], bits[-1]
        for bit in path:
            current[bit] = current.get(bit) or {}
            current = current[bit]
        # Now deal with $type suffixes:
        if _types_re.match(lastkey):
            lastkey, lasttype = lastkey.rsplit("$", 2)
            value = {
                "int": int,
                "float": float,
                "empty": lambda v: {},
                "emptylist": lambda v: [],
                "bool": lambda v: v.lower() == "true",
                "none": lambda v: None,
            }.get(lasttype, lambda v: v)(value)
        current[lastkey] = value

    # We handle foo.[0].one, foo.[1].two syntax in a second pass,
    # by iterating through our structure looking for dictionaries
    # where all of the keys are stringified integers
    def replace_integer_keyed_dicts_with_lists(obj):
        if isinstance(obj, dict):
            if obj and all(_int_key_re.match(k) for k in obj):
                return [
                    i[1]
                    for i in sorted(
                        [
                            (
                                int(_int_key_re.match(k).group(1)),
                                replace_integer_keyed_dicts_with_lists(v),
                            )
                            for k, v in obj.items()
                        ]
                    )
                ]
            else:
                return dict(
                    (k, replace_integer_keyed_dicts_with_lists(v))
                    for k, v in obj.items()
                )
        elif isinstance(obj, list):
            return [replace_integer_keyed_dicts_with_lists(v) for v in obj]
        else:
            return obj

    obj = replace_integer_keyed_dicts_with_lists(obj)
    # Handle root units only, e.g. {'$empty': '{}'}
    if list(obj.keys()) == [""]:
        return list(obj.values())[0]
    return obj



"""CLI Config"""

import base64
import os
import re

from datetime import datetime, timezone


class NotSet:
    pass


class ConfigEntry:
    pattern = re.compile(
        r".*\[profile=(?P<profile>.+)]\n"
        r"USERNAME=(?P<username>.*)\n"
        r"PASSWORD=(?P<password>.*)\n"
        r"TOKEN=(?P<token>.*)\n"
        r"TOKEN_EXPIRES=(?P<token_expires>.*)\n"
        r"AGENT_ID=(?P<agent_id>.*)\n"
        r"BASE_URL=(?P<base_url>.*)"
    )

    def __init__(
        self,
        profile: str,
        username: str = None,
        password: str = None,
        token: str = None,
        token_expires: datetime = None,
        agent_id: str = None,
        base_url: str = None,
    ):
        self.profile = profile

        self.username = username or None
        self.password = password or None

        self.token = token or None
        self.token_expires = token_expires or None

        self.agent_id = agent_id or None
        self.base_url = base_url or None

        self.valid = True

    def __repr__(self):
        return f"ConfigEntry <profile={self.profile}, username={self.username}, base_url={self.base_url}>"

    @classmethod
    def from_data(cls, data):
        match = cls.pattern.match(data.strip())

        if match["token_expires"]:
            token_expires = datetime.fromtimestamp(
                float(match["token_expires"]), tz=timezone.utc
            )
        else:
            token_expires = None

        return cls(
            match["profile"],
            match["username"],
            base64.b64decode(match["password"]).decode("utf-8"),
            match["token"],
            token_expires,
            match["agent_id"],
            match["base_url"],
        )

    def format(self):
        password = self.password or ""
        return (
            f"[profile={self.profile or ''}]\n"
            f"USERNAME={self.username or ''}\n"
            f"PASSWORD={base64.b64encode(password.encode('utf-8')).decode('utf-8') or ''}\n"
            f"TOKEN={self.token or ''}\n"
            f"TOKEN_EXPIRES={self.token_expires and self.token_expires.timestamp() or ''}\n"
            f"AGENT_ID={self.agent_id or ''}\n"
            f"BASE_URL={self.base_url or ''}\n"
        )


class ConfigManager:
    directory = os.path.expanduser("~/.doover")
    filepath = os.path.join(directory, "config")

    def __init__(self, current_profile: str = None):
        self.entries = {}
        self.current_profile = current_profile
        self.read()

    @property
    def current(self) -> ConfigEntry:
        return self.entries.get(self.current_profile)

    def create(self, entry: ConfigEntry):
        self.entries[entry.profile] = entry

    def read(self):
        if not os.path.exists(self.filepath):
            return
            # self.parser.error("Config file doesn't exist. Please run `pydoover configure`.")

        with open(self.filepath, "r") as fp:
            contents = fp.read()

        if len(contents) == 0:
            # protect against empty file
            return

        self.parse(contents)

    def parse(self, contents):
        for item in contents.split("\n\n"):
            config = ConfigEntry.from_data(item)
            self.entries[config.profile] = config

    def write(self):
        if not os.path.exists(self.directory):
            os.mkdir(self.directory)

        fmt = self.dump()  # do this here, so we don't write in case something breaks in formatting config
        with open(self.filepath, "w") as fp:
            fp.write(fmt)

    def dump(self):
        return "\n\n".join(e.format() for e in self.entries.values())



import requests
import json
import copy
from datetime import datetime, timezone
from math import floor

from . import json_flatten

# Module-level constants
SECS_IN_DAY = 24 * 60 * 60
CHANNEL_NAME = "ui_state"


class ReportGeneratorDooverDataMixin:
    """
    Mixin class for adding capabilities for retrieving data from the Doover API to a report generator.
    """

    @staticmethod
    def get_data_object(payload_dict, data_object_name):
        """
        Extract a specific data object from the flattened JSON payload.
        """
        reported_state = payload_dict["state"].get("reported", payload_dict["state"])
        reported_flattened = json_flatten.flatten(reported_state)
        flat_object_key = None

        for key in reported_flattened:
            parts = key.split(".")
            if (
                parts
                and parts[-1] == "name"
                and reported_flattened[key].lower() == data_object_name.lower()
            ):
                flat_object_key = ".".join(parts[:-1])
                break
            elif len(parts) > 2 and parts[-2].lower() == data_object_name.lower():
                flat_object_key = ".".join(parts[:-1])
                break
            elif (
                len(parts) > 3
                and parts[-3].lower() == data_object_name.lower()
                and parts[-2] == "currentValue"
            ):
                flat_object_key = ".".join(parts[:-2])
                break

        if flat_object_key is None:
            return None

        relevant = {}
        for key in reported_flattened:
            if flat_object_key in key:
                trimmed = key.replace(flat_object_key, "", 1)
                if trimmed.startswith("."):
                    trimmed = trimmed[1:]
                relevant[trimmed] = reported_flattened[key]
        return json_flatten.unflatten(relevant)

    @staticmethod
    def add_to_message_log(message, name, object):
        """
        Add a field to a message log for a specific object.
        """
        if "payload" in message:
            if "state" in message["payload"]:
                if "children" in message["payload"]["state"]:
                    message["payload"]["state"]["children"][name] = object
                elif "reported" in message["payload"]["state"]:
                    message["payload"]["state"]["reported"][name] = object
                else:
                    message["payload"]["state"][name] = object
            else:
                message["payload"][name] = object
        else:
            message[name] = object

        return message

    @staticmethod
    def get_available_dataseries(payload_dict):
        """
        Return available data series (uiVariables) from the aggregate payload.
        """
        reported_state = payload_dict["state"].get("reported", payload_dict["state"])
        reported_flattened = json_flatten.flatten(reported_state)
        series_names = []

        for key, value in reported_flattened.items():
            if ".type" in key and value == "uiVariable":
                var_type_key = key.replace(".type", ".varType")
                if (
                    var_type_key in reported_flattened
                    and reported_flattened[var_type_key] == "time"
                ):
                    continue  # Skip time variables
                name_key = key.replace(".type", ".name")
                series_names.append(reported_flattened[name_key])

        result = []
        for series_name in series_names:
            result.append(
                ReportGeneratorDooverDataMixin.get_data_object(
                    payload_dict, series_name
                )
            )
        return result

    def get_current_server_time(self, agent_id):
        """
        Get the current server time (in seconds since the epoch) via the API.
        """
        url = f"{self.api_endpoint}/ch/v1/agent/{agent_id}/"
        resp = requests.get(
            url,
            headers={"Authorization": "Token " + str(self.access_token)},
            verify=(not self.test_mode),
        )
        if resp.status_code == 200:
            result = json.loads(resp.text)
            return int(result["current_time"])
        elif resp.status_code == 403:
            raise Exception("Access Denied - Check Token is Current")
        else:
            self.add_to_log(f"Server Error : {resp.status_code}")
            raise Exception("Server Error")

    def get_current_data_aggregate(self, agent_id):
        """
        Get the current data aggregate from the API.
        """
        url = f"{self.api_endpoint}/ch/v1/agent/{agent_id}/{CHANNEL_NAME}/"
        resp = requests.get(
            url,
            headers={"Authorization": "Token " + str(self.access_token)},
            verify=(not self.test_mode),
        )
        if resp.status_code != 200:
            self.add_to_log(f"Server Error : {resp.status_code}")
            raise Exception("Server Error " + str(resp.status_code))
        data = json.loads(resp.text)
        return data["aggregate"]["payload"]

    def _get_data_for_window(self, window_start, window_end, agent_id):
        """
        Retrieve data messages for a specific time window.
        """
        window_start = int(window_start)
        window_end = int(window_end)
        self.add_to_log(f"Getting data between {window_start} - {window_end}")

        url = (
            f"{self.api_endpoint}/ch/v1/agent/{agent_id}/{CHANNEL_NAME}/"
            f"messages/time/{window_start}/{window_end}/"
        )
        resp = requests.get(
            url,
            headers={"Authorization": "Token " + str(self.access_token)},
            verify=(not self.test_mode),
        )
        if resp.status_code != 200:
            self.add_to_log(f"Server Error : {resp.status_code}")
            raise Exception("Server Error " + str(resp.status_code))
        data = json.loads(resp.text)
        msgs = data["messages"]
        msgs.reverse()  # Order messages from earliest to latest
        return msgs

    def retrieve_data(self, period_from, period_to, agent_id):
        """
        Retrieve data messages over the specified period by splitting the time window into full days
        and any remaining fraction.
        """
        curr_server_time = int(self.get_current_server_time(agent_id))
        local_time = int(datetime.now(timezone.utc).timestamp())
        time_diff = curr_server_time - local_time

        start_secs = (
            copy.copy(period_from).replace(tzinfo=self.for_timezone).timestamp()
            + time_diff
        )
        end_secs = (
            copy.copy(period_to).replace(tzinfo=self.for_timezone).timestamp()
            + time_diff
        )

        num_days = (end_secs - start_secs) / SECS_IN_DAY
        self.add_to_log(f"{num_days} days")

        msgs = []
        for day in range(floor(num_days)):
            temp_start = start_secs + day * SECS_IN_DAY
            temp_end = temp_start + SECS_IN_DAY
            msgs.extend(self._get_data_for_window(temp_start, temp_end, agent_id))

        remainder = num_days % 1
        if remainder > 0:
            remainder_secs = int(remainder * SECS_IN_DAY)
            msgs.extend(
                self._get_data_for_window(end_secs - remainder_secs, end_secs, agent_id)
            )
        return msgs



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from . import platform_iface_pb2 as grpc__stubs_dot_platform__iface__pb2

GRPC_GENERATED_VERSION = '1.65.1'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.66.0'
SCHEDULED_RELEASE_DATE = 'August 6, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in grpc_stubs/platform_iface_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class platformIfaceStub(object):
    """The doover device agent service definition.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TestComms = channel.unary_unary(
                '/platform_iface.platformIface/TestComms',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.TestCommsRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.TestCommsResponse.FromString,
                _registered_method=True)
        self.getDI = channel.unary_unary(
                '/platform_iface.platformIface/getDI',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getDIRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getDIResponse.FromString,
                _registered_method=True)
        self.getAI = channel.unary_unary(
                '/platform_iface.platformIface/getAI',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getAIRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getAIResponse.FromString,
                _registered_method=True)
        self.startPulseCounter = channel.unary_stream(
                '/platform_iface.platformIface/startPulseCounter',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.pulseCounterRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.pulseCounterResponse.FromString,
                _registered_method=True)
        self.getDO = channel.unary_unary(
                '/platform_iface.platformIface/getDO',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getDORequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getDOResponse.FromString,
                _registered_method=True)
        self.setDO = channel.unary_unary(
                '/platform_iface.platformIface/setDO',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.setDORequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.setDOResponse.FromString,
                _registered_method=True)
        self.scheduleDO = channel.unary_unary(
                '/platform_iface.platformIface/scheduleDO',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.scheduleDORequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.scheduleDOResponse.FromString,
                _registered_method=True)
        self.getAO = channel.unary_unary(
                '/platform_iface.platformIface/getAO',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getAORequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getAOResponse.FromString,
                _registered_method=True)
        self.setAO = channel.unary_unary(
                '/platform_iface.platformIface/setAO',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.setAORequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.setAOResponse.FromString,
                _registered_method=True)
        self.scheduleAO = channel.unary_unary(
                '/platform_iface.platformIface/scheduleAO',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.scheduleAORequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.scheduleAOResponse.FromString,
                _registered_method=True)
        self.getValue = channel.unary_unary(
                '/platform_iface.platformIface/getValue',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getValueRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getValueResponse.FromString,
                _registered_method=True)
        self.setValue = channel.unary_unary(
                '/platform_iface.platformIface/setValue',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.setValueRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.setValueResponse.FromString,
                _registered_method=True)
        self.getEvents = channel.unary_unary(
                '/platform_iface.platformIface/getEvents',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getEventsRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getEventsResponse.FromString,
                _registered_method=True)
        self.getDIEvents = channel.unary_unary(
                '/platform_iface.platformIface/getDIEvents',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getDIEventsRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getDIEventsResponse.FromString,
                _registered_method=True)
        self.getSystemStatus = channel.unary_unary(
                '/platform_iface.platformIface/getSystemStatus',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getSystemStatusRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getSystemStatusResponse.FromString,
                _registered_method=True)
        self.getInputVoltage = channel.unary_unary(
                '/platform_iface.platformIface/getInputVoltage',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getInputVoltageRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getInputVoltageResponse.FromString,
                _registered_method=True)
        self.getSystemPower = channel.unary_unary(
                '/platform_iface.platformIface/getSystemPower',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getSystemPowerRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getSystemPowerResponse.FromString,
                _registered_method=True)
        self.getTemperature = channel.unary_unary(
                '/platform_iface.platformIface/getTemperature',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getTemperatureRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getTemperatureResponse.FromString,
                _registered_method=True)
        self.getIoTable = channel.unary_unary(
                '/platform_iface.platformIface/getIoTable',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getIoTableRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getIoTableResponse.FromString,
                _registered_method=True)
        self.syncRtcTime = channel.unary_unary(
                '/platform_iface.platformIface/syncRtcTime',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.syncRtcTimeRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.syncRtcTimeResponse.FromString,
                _registered_method=True)
        self.getLocation = channel.unary_unary(
                '/platform_iface.platformIface/getLocation',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getLocationRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getLocationResponse.FromString,
                _registered_method=True)
        self.getShutdownImmunity = channel.unary_unary(
                '/platform_iface.platformIface/getShutdownImmunity',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getShutdownImmunityRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getShutdownImmunityResponse.FromString,
                _registered_method=True)
        self.setShutdownImmunity = channel.unary_unary(
                '/platform_iface.platformIface/setShutdownImmunity',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.setShutdownImmunityRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.setShutdownImmunityResponse.FromString,
                _registered_method=True)
        self.scheduleStartup = channel.unary_unary(
                '/platform_iface.platformIface/scheduleStartup',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.scheduleStartupRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.scheduleStartupResponse.FromString,
                _registered_method=True)
        self.scheduleShutdown = channel.unary_unary(
                '/platform_iface.platformIface/scheduleShutdown',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.scheduleShutdownRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.scheduleShutdownResponse.FromString,
                _registered_method=True)
        self.reboot = channel.unary_unary(
                '/platform_iface.platformIface/reboot',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.rebootRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.rebootResponse.FromString,
                _registered_method=True)
        self.shutdown = channel.unary_unary(
                '/platform_iface.platformIface/shutdown',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.shutdownRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.shutdownResponse.FromString,
                _registered_method=True)
        self.loadFirmware = channel.unary_unary(
                '/platform_iface.platformIface/loadFirmware',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.loadFirmwareRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.loadFirmwareResponse.FromString,
                _registered_method=True)
        self.loadBootloader = channel.unary_unary(
                '/platform_iface.platformIface/loadBootloader',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.loadBootloaderRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.loadBootloaderResponse.FromString,
                _registered_method=True)
        self.getFirmwareVersion = channel.unary_unary(
                '/platform_iface.platformIface/getFirmwareVersion',
                request_serializer=grpc__stubs_dot_platform__iface__pb2.getFirmwareVersionRequest.SerializeToString,
                response_deserializer=grpc__stubs_dot_platform__iface__pb2.getFirmwareVersionResponse.FromString,
                _registered_method=True)


class platformIfaceServicer(object):
    """The doover device agent service definition.
    """

    def TestComms(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getDI(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getAI(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def startPulseCounter(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getDO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setDO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def scheduleDO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getAO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setAO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def scheduleAO(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getValue(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setValue(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getEvents(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getDIEvents(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getSystemStatus(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getInputVoltage(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getSystemPower(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getTemperature(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getIoTable(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def syncRtcTime(self, request, context):
        """rpc getRtcTime (getRtcTimeRequest) returns (getRtcTimeResponse) {}
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getLocation(self, request, context):
        """rpc getUptime (getUptimeRequest) returns (getUptimeResponse) {}
        rpc getSystemInfo (getSystemInfoRequest) returns (getSystemInfoResponse) {}
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getShutdownImmunity(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setShutdownImmunity(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def scheduleStartup(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def scheduleShutdown(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def reboot(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def shutdown(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def loadFirmware(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def loadBootloader(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getFirmwareVersion(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_platformIfaceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TestComms': grpc.unary_unary_rpc_method_handler(
                    servicer.TestComms,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.TestCommsRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.TestCommsResponse.SerializeToString,
            ),
            'getDI': grpc.unary_unary_rpc_method_handler(
                    servicer.getDI,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getDIRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getDIResponse.SerializeToString,
            ),
            'getAI': grpc.unary_unary_rpc_method_handler(
                    servicer.getAI,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getAIRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getAIResponse.SerializeToString,
            ),
            'startPulseCounter': grpc.unary_stream_rpc_method_handler(
                    servicer.startPulseCounter,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.pulseCounterRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.pulseCounterResponse.SerializeToString,
            ),
            'getDO': grpc.unary_unary_rpc_method_handler(
                    servicer.getDO,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getDORequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getDOResponse.SerializeToString,
            ),
            'setDO': grpc.unary_unary_rpc_method_handler(
                    servicer.setDO,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.setDORequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.setDOResponse.SerializeToString,
            ),
            'scheduleDO': grpc.unary_unary_rpc_method_handler(
                    servicer.scheduleDO,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.scheduleDORequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.scheduleDOResponse.SerializeToString,
            ),
            'getAO': grpc.unary_unary_rpc_method_handler(
                    servicer.getAO,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getAORequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getAOResponse.SerializeToString,
            ),
            'setAO': grpc.unary_unary_rpc_method_handler(
                    servicer.setAO,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.setAORequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.setAOResponse.SerializeToString,
            ),
            'scheduleAO': grpc.unary_unary_rpc_method_handler(
                    servicer.scheduleAO,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.scheduleAORequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.scheduleAOResponse.SerializeToString,
            ),
            'getValue': grpc.unary_unary_rpc_method_handler(
                    servicer.getValue,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getValueRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getValueResponse.SerializeToString,
            ),
            'setValue': grpc.unary_unary_rpc_method_handler(
                    servicer.setValue,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.setValueRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.setValueResponse.SerializeToString,
            ),
            'getEvents': grpc.unary_unary_rpc_method_handler(
                    servicer.getEvents,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getEventsRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getEventsResponse.SerializeToString,
            ),
            'getDIEvents': grpc.unary_unary_rpc_method_handler(
                    servicer.getDIEvents,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getDIEventsRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getDIEventsResponse.SerializeToString,
            ),
            'getSystemStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.getSystemStatus,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getSystemStatusRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getSystemStatusResponse.SerializeToString,
            ),
            'getInputVoltage': grpc.unary_unary_rpc_method_handler(
                    servicer.getInputVoltage,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getInputVoltageRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getInputVoltageResponse.SerializeToString,
            ),
            'getSystemPower': grpc.unary_unary_rpc_method_handler(
                    servicer.getSystemPower,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getSystemPowerRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getSystemPowerResponse.SerializeToString,
            ),
            'getTemperature': grpc.unary_unary_rpc_method_handler(
                    servicer.getTemperature,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getTemperatureRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getTemperatureResponse.SerializeToString,
            ),
            'getIoTable': grpc.unary_unary_rpc_method_handler(
                    servicer.getIoTable,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getIoTableRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getIoTableResponse.SerializeToString,
            ),
            'syncRtcTime': grpc.unary_unary_rpc_method_handler(
                    servicer.syncRtcTime,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.syncRtcTimeRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.syncRtcTimeResponse.SerializeToString,
            ),
            'getLocation': grpc.unary_unary_rpc_method_handler(
                    servicer.getLocation,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getLocationRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getLocationResponse.SerializeToString,
            ),
            'getShutdownImmunity': grpc.unary_unary_rpc_method_handler(
                    servicer.getShutdownImmunity,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getShutdownImmunityRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getShutdownImmunityResponse.SerializeToString,
            ),
            'setShutdownImmunity': grpc.unary_unary_rpc_method_handler(
                    servicer.setShutdownImmunity,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.setShutdownImmunityRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.setShutdownImmunityResponse.SerializeToString,
            ),
            'scheduleStartup': grpc.unary_unary_rpc_method_handler(
                    servicer.scheduleStartup,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.scheduleStartupRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.scheduleStartupResponse.SerializeToString,
            ),
            'scheduleShutdown': grpc.unary_unary_rpc_method_handler(
                    servicer.scheduleShutdown,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.scheduleShutdownRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.scheduleShutdownResponse.SerializeToString,
            ),
            'reboot': grpc.unary_unary_rpc_method_handler(
                    servicer.reboot,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.rebootRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.rebootResponse.SerializeToString,
            ),
            'shutdown': grpc.unary_unary_rpc_method_handler(
                    servicer.shutdown,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.shutdownRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.shutdownResponse.SerializeToString,
            ),
            'loadFirmware': grpc.unary_unary_rpc_method_handler(
                    servicer.loadFirmware,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.loadFirmwareRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.loadFirmwareResponse.SerializeToString,
            ),
            'loadBootloader': grpc.unary_unary_rpc_method_handler(
                    servicer.loadBootloader,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.loadBootloaderRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.loadBootloaderResponse.SerializeToString,
            ),
            'getFirmwareVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.getFirmwareVersion,
                    request_deserializer=grpc__stubs_dot_platform__iface__pb2.getFirmwareVersionRequest.FromString,
                    response_serializer=grpc__stubs_dot_platform__iface__pb2.getFirmwareVersionResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'platform_iface.platformIface', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('platform_iface.platformIface', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class platformIface(object):
    """The doover device agent service definition.
    """

    @staticmethod
    def TestComms(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/TestComms',
            grpc__stubs_dot_platform__iface__pb2.TestCommsRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.TestCommsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getDI(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getDI',
            grpc__stubs_dot_platform__iface__pb2.getDIRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getDIResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getAI(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getAI',
            grpc__stubs_dot_platform__iface__pb2.getAIRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getAIResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def startPulseCounter(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/platform_iface.platformIface/startPulseCounter',
            grpc__stubs_dot_platform__iface__pb2.pulseCounterRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.pulseCounterResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getDO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getDO',
            grpc__stubs_dot_platform__iface__pb2.getDORequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getDOResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def setDO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/setDO',
            grpc__stubs_dot_platform__iface__pb2.setDORequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.setDOResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def scheduleDO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/scheduleDO',
            grpc__stubs_dot_platform__iface__pb2.scheduleDORequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.scheduleDOResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getAO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getAO',
            grpc__stubs_dot_platform__iface__pb2.getAORequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getAOResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def setAO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/setAO',
            grpc__stubs_dot_platform__iface__pb2.setAORequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.setAOResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def scheduleAO(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/scheduleAO',
            grpc__stubs_dot_platform__iface__pb2.scheduleAORequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.scheduleAOResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getValue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getValue',
            grpc__stubs_dot_platform__iface__pb2.getValueRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getValueResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def setValue(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/setValue',
            grpc__stubs_dot_platform__iface__pb2.setValueRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.setValueResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getEvents(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getEvents',
            grpc__stubs_dot_platform__iface__pb2.getEventsRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getEventsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getDIEvents(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getDIEvents',
            grpc__stubs_dot_platform__iface__pb2.getDIEventsRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getDIEventsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getSystemStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getSystemStatus',
            grpc__stubs_dot_platform__iface__pb2.getSystemStatusRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getSystemStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getInputVoltage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getInputVoltage',
            grpc__stubs_dot_platform__iface__pb2.getInputVoltageRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getInputVoltageResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getSystemPower(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getSystemPower',
            grpc__stubs_dot_platform__iface__pb2.getSystemPowerRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getSystemPowerResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getTemperature(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getTemperature',
            grpc__stubs_dot_platform__iface__pb2.getTemperatureRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getTemperatureResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getIoTable(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getIoTable',
            grpc__stubs_dot_platform__iface__pb2.getIoTableRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getIoTableResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def syncRtcTime(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/syncRtcTime',
            grpc__stubs_dot_platform__iface__pb2.syncRtcTimeRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.syncRtcTimeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getLocation(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getLocation',
            grpc__stubs_dot_platform__iface__pb2.getLocationRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getLocationResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getShutdownImmunity(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getShutdownImmunity',
            grpc__stubs_dot_platform__iface__pb2.getShutdownImmunityRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getShutdownImmunityResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def setShutdownImmunity(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/setShutdownImmunity',
            grpc__stubs_dot_platform__iface__pb2.setShutdownImmunityRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.setShutdownImmunityResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def scheduleStartup(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/scheduleStartup',
            grpc__stubs_dot_platform__iface__pb2.scheduleStartupRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.scheduleStartupResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def scheduleShutdown(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/scheduleShutdown',
            grpc__stubs_dot_platform__iface__pb2.scheduleShutdownRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.scheduleShutdownResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def reboot(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/reboot',
            grpc__stubs_dot_platform__iface__pb2.rebootRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.rebootResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def shutdown(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/shutdown',
            grpc__stubs_dot_platform__iface__pb2.shutdownRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.shutdownResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def loadFirmware(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/loadFirmware',
            grpc__stubs_dot_platform__iface__pb2.loadFirmwareRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.loadFirmwareResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def loadBootloader(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/loadBootloader',
            grpc__stubs_dot_platform__iface__pb2.loadBootloaderRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.loadBootloaderResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def getFirmwareVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/platform_iface.platformIface/getFirmwareVersion',
            grpc__stubs_dot_platform__iface__pb2.getFirmwareVersionRequest.SerializeToString,
            grpc__stubs_dot_platform__iface__pb2.getFirmwareVersionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



import asyncio
import argparse
import json
import os
import logging
import time
from collections import deque

from datetime import datetime
from pathlib import Path
from typing import Any, TYPE_CHECKING, Awaitable, Callable

try:
    from aiohttp.web import Response, Server, ServerRunner, TCPSite
except ImportError:
    RUN_HEALTHCHECK = False
else:
    RUN_HEALTHCHECK = True

from .device_agent import DeviceAgentInterface
from .modbus import ModbusInterface
from .platform import PlatformInterface

from ..ui import UIManager
from ..utils import (
    call_maybe_async,
    get_is_async,
    maybe_async,
    setup_logging,
    apply_diff,
    generate_diff,
)

if TYPE_CHECKING:
    from ..config import Schema

log = logging.getLogger(__name__)

MaybeAsyncMainLoopMethod = Callable[[], Awaitable[None]] | Callable[[], None]

TAG_CLOUD_MAX_AGE = 60 * 60  # 1 hour
TAG_CHANNEL_NAME = "tag_values"


class Application:
    """Base class for a Doover application. All apps will inherit from this class, and override the setup and main_loop methods.

    You generally don't need to worry about initiating parameters to this class as that will be done through `run_app`.

    Examples
    --------

    The following is an incredibly simple example of a Doover application that shows how to initiate this Application class.
    However, in practice, it is suggested to use the template application repository for a more structured, complex scaffold
    for building apps.

    A basic application::

        from pydoover.docker import Application, run_app
        from pydoover.config import Schema

        class MyApp(Application):
            def setup(self):
                self.set_tag("my_app_ready", True)

            def main_loop(self):
                # Your main loop logic here
                pass

        if __name__ == "__main__":
            run_app(MyApp(config=Schema()))


    Attributes
    ----------
    config : Schema
        The configuration schema for the application. See [] for more information about config schemas.
    device_agent : DeviceAgentInterface
        The interface to the Doover Device Agent, which allows the app to communicate with the Doover cloud and other devices.
    platform_iface : PlatformInterface
        The interface to the Doover Platform, which allows the app to interact with the device's hardware.
    modbus_iface : ModbusInterface
        The interface to the Modbus communication protocol, allowing the app to read and write Modbus registers.
    ui_manager : UIManager
        The UI manager for the application, which handles the user interface elements and commands.
    app_key : str
        The application key for the app, used to identify it in the Doover cloud. This is globally unique.
    """

    def __init__(
        self,
        config: "Schema",
        app_key: str = None,
        is_async: bool = None,
        device_agent: DeviceAgentInterface = None,
        platform_iface: PlatformInterface = None,
        modbus_iface: ModbusInterface = None,
        name: str = None,
        test_mode: bool = False,
        config_fp: str = None,
        healthcheck_port: int = None,
    ):
        self.config = config

        if config_fp:
            path = Path(config_fp)
            if not path.exists() or not path.is_file():
                raise RuntimeError(f"Config file {config_fp} does not exist")
            self._config_fp = path
        else:
            self._config_fp = None

        self.device_agent = device_agent or DeviceAgentInterface(app_key, "", is_async)
        self.platform_iface = platform_iface or PlatformInterface(app_key, "", is_async)
        self.modbus_iface = modbus_iface or ModbusInterface(
            app_key, "", is_async, config
        )

        self.ui_manager = UIManager(
            app_key=app_key,
            client=self.device_agent,
            is_async=is_async,
        )

        self.app_key = app_key
        self.app_display_name = ""

        self._is_async = get_is_async(is_async)
        self._ready = asyncio.Event()
        self._tag_values = {}
        self._shutdown_at = None
        self.force_log_on_shutdown = False

        if name is None:
            self.name = self.__class__.__name__
        else:
            self.name = name

        self.loop_target_period = 1
        self._error_wait_period = 10

        self._last_interval_time: float | None = None
        self._last_loop_time_warning: float | None = None
        self._loop_times = deque(maxlen=20)

        self._test_next_event = asyncio.Event()
        self._test_next_loop_done = asyncio.Event()
        self.test_mode = test_mode

        self._is_healthy = False
        self._healthcheck_port = healthcheck_port

    async def _handle_healthcheck(self, _request):
        if self._is_healthy:
            return Response(text="OK", status=200)
        else:
            return Response(text="ERROR", status=503)

    async def _on_deployment_config_update(self, _, config: dict[str, Any]):
        try:
            app_config = config["applications"][self.app_key]
        except KeyError:
            log.warning(
                f"Application key {self.app_key} not found in deployment config"
            )
            app_config = {}

        self.device_agent.agent_id = app_config.get("AGENT_ID")
        log.info(f"Agent ID set: {self.device_agent.agent_id}")

        self.app_display_name = app_config.get("APP_DISPLAY_NAME", "")
        log.info(f"Application display name set: {self.app_display_name}")

        log.info(f"Deployment Config Updated: {app_config}")
        self.config._inject_deployment_config(app_config)

    async def __aenter__(self):
        # any more setup here...
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if isinstance(exc_type, (KeyboardInterrupt, asyncio.CancelledError)):
            log.info("Exiting app manager")
        else:
            log.error(f"Error in main loop: {exc_val}", exc_info=exc_tb)

        self._ready.clear()
        await self.close()

    async def next(self):
        """Increment a main loop iteration. This is only available in test mode.

        Normally, the main loop runs in an infinite cycle every `loop_target_period` seconds.

        During testing, it is helpful to be able to control the flow of the main loop, so this method allows you to
        increment the main loop iteration manually. Simply call this method to run the next iteration of the main loop.

        Examples
        --------

        A simple example::

            from pydoover.docker import Application, run_app
            from pydoover.config import Schema

            async def test_app():
                app = MyApp(config=Schema(), test_mode=True)
                asyncio.create_task(run_app(app, start=False))

                # wait for app to start
                await app.wait_until_ready()

                # increment the main loop once
                await app.next()

        Raises
        -------
        RuntimeError
            If this method is called when the app is not in test mode. This method is only available in test mode.
        """
        if not self.test_mode:
            raise RuntimeError("You can only use `app.next()` in test mode.")

        self._test_next_event.set()
        self._test_next_loop_done.clear()
        await self._test_next_loop_done.wait()

    async def _run(self):
        if RUN_HEALTHCHECK:
            try:
                log.info(
                    f"Starting healthcheck server on http://127.0.0.1:{self._healthcheck_port}"
                )
                server = Server(self._handle_healthcheck)
                runner = ServerRunner(server)
                await runner.setup()
                site = TCPSite(runner, "127.0.0.1", self._healthcheck_port)
                await site.start()
            except Exception as e:
                log.error(f"Error starting healthcheck server: {e}", exc_info=e)
        else:
            log.info("`aiohttp` not installed, skipping healthcheck server.")

        if self._config_fp is not None:
            data = json.loads(self._config_fp.read_text())
            self.config._inject_deployment_config(data)
        else:
            self.device_agent.add_subscription(
                "deployment_config", self._on_deployment_config_update
            )
            await self.device_agent.wait_for_channels_sync_async(["deployment_config"])

        await self.modbus_iface.setup()

        try:
            await self._setup()
            ## Don't run setup in executor, to give the user the
            # ability to set up async stuff in the setup function
            await call_maybe_async(self.setup, in_executor=False)
        except Exception as e:
            log.error(f"Error in setup function: {e}", exc_info=e)
            log.warning(
                f"\n\nWaiting {self._error_wait_period} seconds before restarting app\n\n"
            )
            await asyncio.sleep(self._error_wait_period)
            return

        self.ui_manager.start_comms()
        await self.ui_manager.await_comms_sync(15)
        self._ready.set()

        try:
            shutdown_at = self._tag_values["shutdown_at"]
        except KeyError:
            pass
        else:
            await self._check_shutdown_at(shutdown_at)

        ## allow for other async tasks to run between setup and loop
        await asyncio.sleep(0.2)

        while True:
            if self.test_mode:
                await self._test_next_event.wait()
                self._test_next_event.clear()  # clear it for the next iteration...

            try:
                await self._main_loop()
                await call_maybe_async(self.main_loop)
            except Exception as e:
                log.error(f"Error in loop function: {e}", exc_info=e)
                log.warning(
                    f"\n\n\nWaiting {self._error_wait_period} seconds before restarting app\n\n"
                )
                self._is_healthy = False
                await asyncio.sleep(self._error_wait_period)
                break
            else:
                if self.test_mode is False:
                    # slow down the loop in live mode.
                    # await asyncio.sleep(self.loop_target_period)
                    await self.wait_for_interval(self.loop_target_period)
                else:
                    # allow other async tasks to run if the user has done a doozy and chained a whole heap of .next()s
                    await asyncio.sleep(0.01)

                self._is_healthy = True

            if self.test_mode is True:
                # signal that the loop is done.
                self._test_next_loop_done.set()

    async def wait_for_interval(self, target_time: float):
        """
        Waits for the necessary amount of time to maintain a consistent interval
        of `target_time` seconds between calls to this method.
        """

        current_time = time.time()
        if self._last_interval_time is None:
            self._last_interval_time = current_time
            ## Wait for half the target time on the first call
            await asyncio.sleep(target_time / 2)
            return

        elapsed = current_time - self._last_interval_time
        await self._assess_loop_time(
            elapsed, target_time
        )  ## This will display a warning if the loop is running slower than target
        elapsed = current_time - self._last_interval_time
        remaining = target_time - elapsed
        log.debug(f"Last loop time: {elapsed}, target_time: {target_time}")
        if remaining > 0:
            log.debug(f"Sleeping for {remaining} seconds to maintain target loop time")
            await asyncio.sleep(remaining)
        self._last_interval_time = time.time()

    async def _assess_loop_time(self, last_loop_time: float, target_time: float):
        """
        Assess the loop time and adjust the target time if necessary.
        """
        self._loop_times.append(last_loop_time)
        average_loop_time = sum(self._loop_times) / len(self._loop_times)
        log.debug(f"Average loop time: {average_loop_time}, target_time: {target_time}")

        ## If the loop time is greater than 20% above the target time, display a warning every 6 seconds or so
        if average_loop_time > (target_time * 1.2):
            if (
                not hasattr(self, "_last_loop_time_warning")
                or self._last_loop_time_warning is None
            ):
                self._last_loop_time_warning = time.time()
            elif time.time() - self._last_loop_time_warning > 6:
                log.warning(
                    f"Loop is running slower than target. Average loop time: {average_loop_time}, target_time: {target_time}"
                )
                self._last_loop_time_warning = time.time()

    async def close(self):
        log.info(
            "\n########################################"
            "\n\nClosing app manager...\n\n"
            "########################################\n"
        )

        await self.device_agent.close()
        await self.platform_iface.close()
        await self.modbus_iface.close()

        for task in asyncio.all_tasks():
            task.cancel()

    @property
    def is_ready(self) -> bool:
        """Check if the application is ready.

        The application is ready when all initialization tasks have completed and the UI is set up.
        In practice, this means your `setup` method has completed and the application is connected to the cloud.

        Returns
        -------
        bool
            True if the application is ready, False otherwise.
        """
        return self._ready.is_set()

    async def wait_until_ready(self):
        """Wait until the application is ready.

        This method waits (blocks) the current loop until the application is ready.
        """
        await self._ready.wait()

    ## Agent Interface Functions (DDA)

    def get_is_dda_available(self):
        return self.device_agent.get_is_dda_available()

    def get_is_dda_online(self):
        return self.device_agent.get_is_dda_online()

    def get_has_dda_been_online(self):
        return self.device_agent.get_has_dda_been_online()

    def subscribe_to_channel(
        self,
        channel_name: str,
        callback: (
            Callable[[str, dict[str, Any]], Awaitable[Any]]
            | Callable[[str, dict[str, Any]], Any]
        ),
    ):
        return self.device_agent.add_subscription(channel_name, callback)

    def publish_to_channel(self, channel_name: str, data: str | dict[str, Any]):
        return self.device_agent.publish_to_channel(channel_name, data)

    def get_channel_aggregate(self, channel_name: str):
        return self.device_agent.get_channel_aggregate(channel_name)

    ## UI Manager Functions

    def set_ui_elements(self, elements):
        return self.ui_manager.set_children(elements)

    def get_command(self, name):
        return self.ui_manager.get_command(name)

    def coerce_command(self, name, value):
        return self.ui_manager.coerce_command(name, value)

    def record_critical_value(self, name, value):
        return self.ui_manager.record_critical_value(name, value)

    def set_ui_status_icon(self, icon):
        return self.ui_manager.set_status_icon(icon)

    def start_ui_comms(self):
        return self.ui_manager.start_comms()

    async def await_ui_comms_sync(self, timeout=10):
        log.debug("Awaiting UI comms sync")
        result = await self.ui_manager.await_comms_sync(timeout=timeout)
        if result is False:
            log.warning("UI comms sync timed out")
        else:
            log.debug("UI comms sync complete")
        return result

    def set_ui(self, ui):
        self.ui_manager.set_children(ui)

    async def _update_ui(self, force_log: bool = False):
        await self.ui_manager.handle_comms_async(force_log)

    ## Platform Interface Functions

    def get_di(self, di):
        return self.platform_iface.get_di(di)

    def get_ai(self, ai):
        return self.platform_iface.get_ai(ai)

    def get_do(self, do):
        return self.platform_iface.get_do(do)

    def set_do(self, do, value):
        return self.platform_iface.set_do(do, value)

    def schedule_do(self, do, value, delay_secs):
        return self.platform_iface.schedule_do(do, value, delay_secs)

    def get_ao(self, ao):
        return self.platform_iface.get_ao(ao)

    def set_ao(self, ao, value):
        return self.platform_iface.set_ao(ao, value)

    def schedule_ao(self, ao, value, delay_secs):
        return self.platform_iface.schedule_ao(ao, value, delay_secs)

    ## Modbus Interface Functions

    def read_modbus_registers(
        self, address, count, register_type, modbus_id=None, bus_id=None
    ):
        return self.modbus_iface.read_registers(
            bus_id=bus_id,
            modbus_id=modbus_id,
            start_address=address,
            num_registers=count,
            register_type=register_type,
        )

    def write_modbus_registers(
        self, address, values, register_type, modbus_id=None, bus_id=None
    ):
        return self.modbus_iface.write_registers(
            bus_id=bus_id,
            modbus_id=modbus_id,
            start_address=address,
            values=values,
            register_type=register_type,
        )

    def add_new_modbus_read_subscription(
        self,
        address,
        count,
        register_type,
        callback,
        poll_secs=None,
        modbus_id=None,
        bus_id=None,
    ):
        return self.modbus_iface.add_read_register_subscription(
            bus_id=bus_id,
            modbus_id=modbus_id,
            start_address=address,
            num_registers=count,
            register_type=register_type,
            poll_secs=poll_secs,
            callback=callback,
        )

    # state

    @property
    def _shutdown_requested(self):
        try:
            return self._tag_values["shutdown_requested"]
        except (KeyError, TypeError):
            return False

    async def _on_tag_update(self, _, tag_values: dict[str, Any]):
        self._tag_values = tag_values or {}

        try:
            shutdown_at = tag_values["shutdown_at"]
        except (KeyError, TypeError):
            pass
        else:
            await self._check_shutdown_at(shutdown_at)

    async def _check_shutdown_at(self, shutdown_at):
        if not self.is_ready:
            log.info("Ignoring check shutdown request, app not ready yet.")
            return

        dt = datetime.fromtimestamp(shutdown_at)
        if self._shutdown_at is None or (
            dt > self._shutdown_at and dt > datetime.now()
        ):
            # shutdown should be in the future and not already scheduled
            log.info(f"Shutdown scheduled at {dt.strftime('%Y-%m-%d %H:%M:%S')}")
            self._shutdown_at = dt
            await call_maybe_async(self.on_shutdown_at, dt)

    def get_tag(
        self, tag_key: str, app_key: str = None, default: Any = None
    ) -> Any | None:
        """Get a tag value for a specific app.

        If you want to get a global tag, use :meth:`get_global_tag` instead.

        Examples
        --------

        >>> tag_value = self.get_tag("other_tag", "some-other-app-1234")
        >>> print(f"other-tag is {tag_value} for app some-other-app-1234")

        >>> tag_value = self.get_tag("my_tag")
        >>> print(f"my-tag is {tag_value} for current app {self.app_key}")


        Parameters
        ----------
        tag_key: str
            The tag to fetch.
        app_key: str, optional
            The app key to get the tag for. This defaults to the current app.
        default: Any, optional
            The default value to return if the tag does not exist. Defaults to None.


        Returns
        -------
        Any
            The value of the tag, or None if the tag does not exist.
        """
        try:
            if app_key is None:
                app_key = self.app_key
            return self._tag_values[app_key][tag_key]
        except (KeyError, TypeError):
            log.debug(f"Tag {tag_key} not found in current tags")
            return default

    def get_global_tag(self, tag_key: str, default: Any = None) -> Any | None:
        """Get a global tag value.

        Global tags are tags that are not specific to an app, but are shared across all apps.

        Warnings
        --------
        Due to namespacing concerns, it's best practice to use global tags sparingly and only for values that are truly global in nature.
        For example, you might use a global tag for a shutdown request or a system-wide status indicator.
        If you need to get a tag for a specific app, use :meth:`get_tag` instead.

        Examples
        --------
        >>> is_flag_set = self.get_global_tag("my_global_flag")
        >>> print(f"Global flag my_global_flag is set to {is_flag_set}")

        Parameters
        ----------
        tag_key: str
            The global tag to fetch.
        default: Any, optional
            The default value to return if the tag does not exist. Defaults to None.

        Returns
        -------
        Any
            The value of the global tag, or None if the tag does not exist.
        """
        try:
            return self._tag_values[tag_key]
        except (KeyError, TypeError):
            log.debug(f"Global tag {tag_key} not found in current tags")
            return default

    @maybe_async()
    def set_tag(
        self,
        tag_key: str,
        value: Any,
        app_key: str = None,
        only_if_changed: bool = True,
    ) -> None:
        """Set a tag value.

        This method sets a tag value for a specific app. If you want to set a global tag, use :meth:`set_global_tag` instead.

        .. note:: This method can be called in both sync and asynchronous contexts.

        Examples
        --------
        >>> self.set_tag("my_tag", "my_value")
        >>> self.set_tag("other_tag", "other_value", app_key="some-other-app-1234")

        Parameters
        ----------
        tag_key: str
            The tag to set.
        value: Any
            The value to set the tag to.
        app_key: str, optional
            The app key to set the tag for. This defaults to the current app's key.
        only_if_changed: bool, optional
            If True, the tag will only be set if the value is different from the current value. Defaults to True.
        """
        self._do_set_tags(
            {tag_key: value}, app_key=app_key, only_if_changed=only_if_changed
        )

    async def set_tag_async(
        self,
        tag_key: str,
        value: Any,
        app_key: str = None,
        only_if_changed: bool = True,
    ) -> None:
        await self._do_set_tags_async(
            {tag_key: value}, app_key=app_key, only_if_changed=only_if_changed
        )

    @maybe_async()
    def set_tags(
        self, tags: dict[str, Any], app_key: str = None, only_if_changed: bool = True
    ) -> None:
        """Set multiple tags at once."""
        self._do_set_tags(tags, app_key=app_key, only_if_changed=only_if_changed)

    async def set_tags_async(
        self, tags: dict[str, Any], app_key: str = None, only_if_changed: bool = True
    ) -> None:
        await self._do_set_tags_async(
            tags, app_key=app_key, only_if_changed=only_if_changed
        )

    @maybe_async()
    def set_global_tag(
        self, tag_key: str, value: Any, only_if_changed: bool = True
    ) -> None:
        """Set a global tag value.

        As in :meth:`get_global_tag`, global tags are not specific to an app, but are shared across all apps and should be used sparingly as such.

        .. note:: This method can be called in both sync and asynchronous contexts.

        Examples
        --------
        >>> self.set_global_tag("my_global_flag", True)
        >>> self.set_global_tag("system_status", "operational")

        Parameters
        ----------
        tag_key: str
            The global tag to set.
        value: Any
            The value to set the global tag to.
        only_if_changed: bool, optional
            If True, the tag will only be set if the value is different from the current value. Defaults to True.
        """
        self._do_set_tags(
            {tag_key: value},
            app_key=None,
            is_global=True,
            only_if_changed=only_if_changed,
        )

    async def set_global_tag_async(
        self, tag_key: str, value: Any, only_if_changed: bool = True
    ) -> None:
        """Set a global tag value asynchronously. This is a convenience method for setting global tags."""
        await self._do_set_tags_async(
            {tag_key: value},
            app_key=None,
            is_global=True,
            only_if_changed=only_if_changed,
        )

    def _do_set_tags(
        self,
        tags: dict[str, Any],
        app_key: str | None,
        is_global: bool = False,
        only_if_changed: bool = True,
    ):
        if is_global:
            data = tags
        else:
            if app_key is None:
                app_key = self.app_key
            data = {app_key: tags}

        if only_if_changed:
            diff = generate_diff(self._tag_values, data, do_delete=False)
            if len(diff) == 0:
                return

        apply_diff(self._tag_values, data, clone=False)
        self.device_agent.publish_to_channel(
            TAG_CHANNEL_NAME, data, max_age=TAG_CLOUD_MAX_AGE, record_log=True
        )

    async def _do_set_tags_async(
        self,
        tags: dict[str, Any],
        app_key: str | None,
        is_global: bool = False,
        only_if_changed: bool = True,
    ):
        if is_global:
            data = tags
        else:
            if app_key is None:
                app_key = self.app_key
            data = {app_key: tags}

        if only_if_changed:
            diff = generate_diff(self._tag_values, data, do_delete=False)
            if len(diff) == 0:
                return

        apply_diff(self._tag_values, data, clone=False)
        await self.device_agent.publish_to_channel_async(
            TAG_CHANNEL_NAME, data, max_age=TAG_CLOUD_MAX_AGE, record_log=True
        )

    ## Power Manager Functions

    @maybe_async()
    def request_shutdown(self) -> None:
        """Request a system shutdown

        .. note:: This method can be called in both sync and asynchronous contexts.
        """
        log.info("Requesting shutdown")
        self.set_tag("shutdown_requested", True)

    async def request_shutdown_async(self) -> None:
        log.info("Requesting shutdown")
        await self.set_tag_async("shutdown_requested", True)

    async def on_shutdown_at(self, dt: datetime) -> None:
        """Callback for when a shutdown is scheduled.

        See [https://docs.doover.com/docker/shutdown-behaviour] for a detailed explanation of the shutdown behaviour.

        This method is called when a shutdown is scheduled, and can be overridden by an application to perform
        specific actions before the imminent system shutdown.

        By default, this method does nothing.

        Examples
        --------

        Simple logging example::

            class MyApplication(Application):
                # setup, main_loop, etc...

                async def on_shutdown_at(self, dt: datetime):
                    log.info(f"Shutdown scheduled at {dt}. Performing cleanup...")


        Parameters
        ----------
        dt : datetime
            The datetime when the shutdown is scheduled.
        """
        if self.force_log_on_shutdown:
            await self._update_ui(force_log=True)

    async def check_can_shutdown(self) -> bool:
        """Check if the application can shutdown.

        This method is called when the application is requested to shutdown,
        and should be overridden by an application if specific logic is required when a shutdown is requested.

        See [https://docs.doover.com/docker/shutdown-behaviour] for a detailed explanation of the shutdown behaviour.

        This must be implemented as an asynchronous function, take no parameters and return a boolean value.

        A return value of `True` indicates that the application can shutdown, while `False` indicates that it cannot.

        By default, this method always returns `True`, meaning the application can shutdown without any checks.

        Examples
        --------
        Simple example that checks if Digital Output 0 (maybe an engine or fan) is Low before returning True::

            class MyApplication(Application):
                # setup, main_loop, etc...

                async def check_can_shutdown(self) -> bool:
                    if await self.platform_iface.get_do(0) == 0:
                        log.info("Digital Output 0 is Low. Can shutdown.")
                        return True
                    else:
                        log.warning("Digital Output 0 is High. Cannot shutdown.")
                        return False

        """
        return True

    ## App Functions

    async def _setup(self):
        log.info(f"Setting up internal app: {self.name}")
        self.ui_manager.register_callbacks(self)
        self.ui_manager.set_display_name(self.app_display_name)
        self.device_agent.add_subscription(TAG_CHANNEL_NAME, self._on_tag_update)
        await self.ui_manager.clear_ui_async()

    async def _main_loop(self):
        log.debug(f"Running internal main_loop: {self.name}")
        if self._shutdown_requested:
            try:
                resp = await self.check_can_shutdown()
            except Exception as e:
                log.error(
                    f"Error checking if we can shutdown: {e}. Assuming False.",
                    exc_info=e,
                )
                resp = False

            await self.set_tag_async("shutdown_check_ok", resp)

        await self._update_ui()

    async def setup(self):
        """The main setup function for the application.

        Your application should override this method to perform any setup tasks that need to be done before the main loop starts.

        Generally, that involves setting up UI, registering callbacks, starting state machines, etc.

        This function can be asynchronous or synchronous, depending on your needs.

        You do **not** need to call `super()` inside your setup method; this function does nothing by default.
        """
        return NotImplemented

    async def main_loop(self):
        """The main loop function for the application.

        Your application should override this method to perform the main logic of your application.

        Generally, this involves running and checking any state machines, setting tags, reading sensors, etc. depending on your application.

        This function is called in a continuous loop, so it should generally not perform any long blocking calls, instead deferring to
        checking if a result is ready to be processed in a future loop.

        You can control the speed at which this loop runs by setting the `loop_target_period` attribute of the application instance.
        By default, this is set to a target invocation period of 1 second.

        This function can be asynchronous or synchronous, depending on your needs.

        You do **not** need to call `super()` inside your setup method; this function does nothing by default.
        """
        return NotImplemented


def parse_args():
    parser = argparse.ArgumentParser(description="Doover Docker App Manager")

    parser.add_argument("--app-key", type=str, default=None, help="App Key")
    parser.add_argument(
        "--remote-dev", type=str, default=None, help="Remote device URI"
    )
    parser.add_argument(
        "--dda-uri", type=str, default=None, help="Doover Device Agent URI"
    )
    parser.add_argument(
        "--plt-uri", type=str, default="localhost:50053", help="Platform Interface URI"
    )
    parser.add_argument(
        "--modbus-uri", type=str, default="localhost:50054", help="Modbus Interface URI"
    )
    parser.add_argument(
        "--config-fp",
        type=str,
        default=None,
        help="Config file path to override app config",
    )
    parser.add_argument(
        "--healthcheck-port",
        type=int,
        default=None,
        help="Port for the healthcheck server (default: 49200). This must be overidden per-app to avoid conflicts.",
    )
    parser.add_argument(
        "--debug", action="store_true", default=False, help="Debug Mode"
    )

    args = parser.parse_args()

    app_key = args.app_key or os.environ.get("APP_KEY")
    dda_uri = args.dda_uri or os.environ.get("DDA_URI") or "localhost:50051"
    plt_uri = args.plt_uri or os.environ.get("PLT_URI") or "localhost:50053"
    modbus_uri = args.modbus_uri or os.environ.get("MODBUS_URI") or "localhost:50054"
    healthcheck_port = int(
        args.healthcheck_port or os.environ.get("HEALTHCHECK_PORT") or 49200
    )
    config_fp = args.config_fp or os.environ.get("CONFIG_FP")

    remote_dev = args.remote_dev or os.environ.get("REMOTE_DEV")
    if remote_dev is not None:
        dda_uri = dda_uri.replace("localhost", remote_dev)
        plt_uri = plt_uri.replace("localhost", remote_dev)
        modbus_uri = modbus_uri.replace("localhost", remote_dev)

    debug = args.debug or os.environ.get("DEBUG") == "1"
    return (
        app_key,
        dda_uri,
        plt_uri,
        modbus_uri,
        remote_dev,
        config_fp,
        debug,
        healthcheck_port,
    )


def run_app(app: Application, start: bool = True):
    """Run the application.

    This function initializes the application, sets up the interfaces, and runs the main loop.
    If `start` is True, it will run the application in a blocking manner, otherwise it will return an async runner function.
    This is useful for testing or when you want to run the application in an event loop without blocking the main thread, but not recommended for production use.

    Examples
    --------

    The general recommended structure for starting applications in the `__init__.py` file::

        from pydoover.docker import run_app

        from .application import SampleApplication
        from .app_config import SampleConfig

        def main():
            run_app(SampleApplication(config=SampleConfig()))


    Parameters
    ----------
    app : Application
        The application instance to run.
    start : bool, optional
        If True, the application will run in a blocking manner. If False, it will return an async runner function.
        Defaults to True.
    """
    (
        app_key,
        dda_uri,
        plt_uri,
        modbus_uri,
        remote_dev,
        config_fp,
        debug,
        healthcheck_port,
    ) = parse_args()

    user_is_async = asyncio.iscoroutinefunction(
        app.setup
    ) or asyncio.iscoroutinefunction(app.main_loop)
    is_async = get_is_async(user_is_async)
    setup_logging(debug)

    for inst in (
        app,
        app.platform_iface,
        app.modbus_iface,
        app.device_agent,
        app.ui_manager,
    ):
        inst.app_key = app_key
        inst._is_async = is_async

    app.platform_iface.uri = plt_uri
    app.modbus_iface.uri = modbus_uri
    app.device_agent.uri = dda_uri
    app._config_fp = config_fp and Path(config_fp)
    app._healthcheck_port = healthcheck_port

    async def runner():
        async with app:
            await app._run()

    if start:
        try:
            asyncio.run(runner())
        except KeyboardInterrupt:
            pass
    else:
        return runner()


def run_app2(
    app_cls: type[Application],
    config: "Schema",
    dda_iface_cls: type[DeviceAgentInterface] = DeviceAgentInterface,
    plt_iface_cls: type[PlatformInterface] = PlatformInterface,
    mb_iface_cls: type[ModbusInterface] = ModbusInterface,
):
    (
        app_key,
        dda_uri,
        plt_uri,
        modbus_uri,
        remote_dev,
        config_fp,
        debug,
        healthcheck_port,
    ) = parse_args()

    user_is_async = asyncio.iscoroutinefunction(
        app_cls.setup
    ) or asyncio.iscoroutinefunction(app_cls.main_loop)
    is_async = get_is_async(user_is_async)
    setup_logging(debug)

    app = app_cls(
        config,
        app_key,
        is_async,
        platform_iface=plt_iface_cls(app_key, plt_uri, is_async),
        modbus_iface=mb_iface_cls(app_key, modbus_uri, is_async, config),
        device_agent=dda_iface_cls(app_key, dda_uri, is_async),
        config_fp=config_fp,
        healthcheck_port=healthcheck_port,
    )

    async def runner():
        async with app:
            await app._run()

    try:
        asyncio.run(runner())
    except KeyboardInterrupt:
        pass



import argparse
import inspect
import traceback

import logging
import os
import sys

from .sub_section import SubSection


class CLI:
    def __init__(self):
        parser = argparse.ArgumentParser(
            prog="pydoover", description="Interact with running gRPC servers."
        )
        parser.set_defaults(callback=parser.print_help)

        self.subparser = parser.add_subparsers(dest="subcommand", title="Subcommands")
        self.added_subsections = []

        # to stop circular imports...
        try:
            # fixme: make a [docker] extra feature / package which processors can choose not to install.
            from ..docker.platform import PlatformInterface
            from ..docker.device_agent import DeviceAgentInterface
            from ..docker.modbus import ModbusInterface
        except ImportError as e:
            print(e)
            print(
                "Docker interfaces not found. GRPC CLI support will not be available."
            )
        else:
            self.add_grpc_subsection(
                SubSection(
                    PlatformInterface,
                    name="platform",
                    description="Interact with a running Platform Interface container",
                )
            )
            self.add_grpc_subsection(
                SubSection(
                    DeviceAgentInterface,
                    name="device_agent",
                    description="Interact with a running Device Agent container",
                )
            )
            self.add_grpc_subsection(
                SubSection(
                    ModbusInterface,
                    name="modbus",
                    description="Interact with a running Modbus Interface container",
                )
            )

        self.args = args = parser.parse_args()

        # remove grcp logging while using cli
        os.environ["GRPC_VERBOSITY"] = "ERROR"
        os.environ["GRPC_TRACE"] = ""
        logging.getLogger().setLevel(logging.ERROR)
        sys.stdout.reconfigure(line_buffering=True)

        try:
            passed_args = {
                k: v
                for k, v in vars(args).items()
                if k in inspect.signature(args.callback).parameters.keys()
            }
            if "kwargs" in inspect.signature(args.callback).parameters.keys():
                passed_args = {k: v for k, v in vars(args).items()}
            args.callback(**passed_args)
        except Exception as e:
            if args.debug:
                traceback.print_exc()
            else:
                print(f"An error occurred: {e}")

    def add_grpc_subsection(self, subsection: SubSection):
        subsection.mount_sub_section(self.subparser)
        self.added_subsections.append(subsection)

    def main(self):
        pass
